#!/usr/bin/python

import netsim
import paxos

class Validator(object):
    __slots__ = ("validate_count")

    def __init__(self):
        self.validate_count = 0

    def validate(self, nodes):
        #~ print "validating state ..."
        self.validate_count += 1
        if self.validate_count % 100000 == 0:
            print self.validate_count
        proposer_accepted_values = set()
        acceptor_accepted_values = set()
        for node_id, node in enumerate(nodes):
            if isinstance(node, paxos.Proposer):
                if node.accepted:
                    proposer_accepted_values.add(node.proposal_value)
                #~ print id, node.votes, node.prepared, node.accepted
            else:
                assert isinstance(node, paxos.Acceptor)
                acceptor_accepted_values.add(node.value)

        assert len(acceptor_accepted_values) == 1
        return len(proposer_accepted_values) <= 1


# executed histories, rejected choices, validated states
# 1a, 2p = 15 35; 5 validated states
# 2a, 2p = 1254 5154; 208 validated states
# 2a, 3p = 427881 2517117; 27328 validated states
# 3a, 1p = 525 1598; 288 validated states
# 3a, 2p = 28980244 212554915; 4962816 validated states; 122m @3GHz P4 (farm5)
# 3a, 3p = no bug found after 70700000 states and 6869m46.431s
# 4a, 1p = 12880 52787; 5760 validated states
# 4a, 2p = Segmentation fault after 5836m34.239s
# 5a, 1p = 2490205 10054935, 1209600 validated states
ACCEPTOR_COUNT = 2
PROPOSER_COUNT = 3
def createNetwork(system):
    acceptors = []
    for i in xrange(ACCEPTOR_COUNT):
        acceptors.append(system.addNode(paxos.Acceptor()))

    for i in xrange(PROPOSER_COUNT):
        value = chr(ord("a") + i)
        system.addNode(paxos.Proposer(acceptors, value))


def main():
    #~ print "%da %dp" % (ACCEPTOR_COUNT, PROPOSER_COUNT)
    validator = Validator()
    errors = netsim.simulateAll(createNetwork, validator.validate)

    print validator.validate_count, "validated states"
    print len(errors), "errors"

    #~ for error in errors:
        #~ print "resimulate ..."
        #~ system = netsim.System()
        #~ system.resimulate(createNetwork, error)

    #~ choices = [0, 0, 7, 7, 7, 2, 2, 8, 8, 9, 9, 9, 9, 2, 2, 8, 7, 7, 7, 7, 7, 7, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    #~ system = netsim.System()
    #~ system.resimulate(createNetwork, choices)
    #~ import sys
    #~ choices = []
    #~ system.resimulate(createNetwork, [])
    #~ while len(system.event_queue) > 0:
        #~ for i, event in enumerate(system.event_queue):
            #~ print "%d: %d -> %d %s" % (i, event.source, event.destination, event.data)
        #~ print "choice? ",
        #~ sys.stdout.flush()
        #~ choice = int(raw_input())
        #~ choices.append(choice)
        #~ system.executeChoice(choice)

    #~ print choices
    #~ if not validator.validate([n.state_machine for n in system.nodes]):
        #~ print "error!"


if __name__ == "__main__":
    #~ import cProfile
    #~ import pstats
    #~ import tempfile
    #~ out = tempfile.NamedTemporaryFile()
    #~ cProfile.run("main()", out.name)
    #~ stats = pstats.Stats(out.name)
    #~ stats.sort_stats("time")
    #~ stats.print_stats()
    # Try to use psyco if possible, which makes things ~5X faster
    try:
        import psyco
        psyco.full()
    except ImportError:
        print "WARNING: Not using psyco"
        pass
    main()
